from __future__ import absolute_import, division, print_function

import datetime
import logging
import re
from importlib import import_module
from itertools import chain
from urllib.parse import urlparse

import numpy
from affine import Affine
from datacube.utils import geometry
from dateutil.parser import parse
from flask import request
from pytz import timezone, utc
from rasterio import MemoryFile
from timezonefinderL import TimezoneFinder

_LOG = logging.getLogger(__name__)


tf = TimezoneFinder(in_memory=True)

# Use metadata time if possible as this is what WMS uses to calculate it's temporal extents
# datacube-core center time accessed through the dataset API is calculated and may
# not agree with the metadata document


def dataset_center_time(dataset):
    center_time = dataset.center_time
    try:
        metadata_time = dataset.metadata_doc['extent']['center_dt']
        center_time = parse(metadata_time)
    except KeyError:
        try:
            metadata_time = dataset.metadata_doc['properties']['dtr:start_datetime']
            center_time = parse(metadata_time)
        except KeyError:
            pass
    return center_time


class NoTimezoneException(Exception):
    pass


def solar_date(dt, tz):
    return dt.astimezone(tz).date()


def local_date(ds, tz=None):
    dt_utc = dataset_center_time(ds)
    if tz:
        return dt_utc.astimezone(tz).date()
    else:
        return dt_utc.astimezone(tz_for_geometry(ds.extent)).date()


def tz_for_dataset(ds):
    return tz_for_geometry(ds.extent)


def tz_for_coord(lon, lat):
    try:
        tzn = tf.timezone_at(lng=lon, lat=lat)
    except:
        print("Woah!")
        raise
    if not tzn:
        raise NoTimezoneException("tz find failed.")
    return timezone(tzn)


def local_solar_date_range(geobox, date):
    tz = tz_for_geometry(geobox.geographic_extent)
    start = datetime.datetime(date.year, date.month, date.day, 0, 0, 0, tzinfo=tz)
    end = datetime.datetime(date.year, date.month, date.day, 23, 59, 59, tzinfo=tz)
    return (start.astimezone(utc), end.astimezone(utc))


def month_date_range(date):
    start = datetime.datetime(date.year, date.month, 1, 0, 0, 0)
    y = date.year
    m = date.month + 1
    if m == 13:
        m = 1
        y = y + 1
    end = datetime.datetime(y, m, 1, 0, 0, 0) - datetime.timedelta(days=1)
    return start, end


def year_date_range(date):
    start = datetime.datetime(date.year, 1, 1, 0, 0, 0)
    end = datetime.datetime(date.year, 12, 31, 23, 59, 59)
    return start, end


def tz_for_geometry(geom):
    crs_geo = geometry.CRS("EPSG:4326")
    geo_geom = geom.to_crs(crs_geo)
    centroid = geo_geom.centroid
    try:
        return tz_for_coord(centroid.coords[0][0], centroid.coords[0][1])
    except NoTimezoneException:
        pass
    for pt in geo_geom.boundary.coords:
        try:
            return tz_for_coord(pt[0], pt[1])
        except NoTimezoneException:
            pass
    offset = round(centroid.coords[0][0] / 15.0)
    return datetime.timezone(datetime.timedelta(hours=offset))


def resp_headers(d):
    from datacube_ows.ows_configuration import get_config
    return get_config().response_headers(d)


def get_function(func):
    """Converts a config entry to a function, if necessary

    :param func: Either a Callable object or a fully qualified function name str, or None
    :return: a Callable object, or None
    """
    if func is not None and not callable(func):
        mod_name, func_name = func.rsplit('.', 1)
        mod = import_module(mod_name)
        func = getattr(mod, func_name)
        assert callable(func)
    return func


def parse_for_base_url(url):
    parsed = urlparse(url)
    parsed = (parsed.netloc + parsed.path).rstrip("/")
    return parsed


def get_service_base_url(allowed_urls, request_url):
    if not isinstance(allowed_urls, list):
        return allowed_urls
    parsed_request_url = parse_for_base_url(request_url)
    parsed_allowed_urls = [parse_for_base_url(u) for u in allowed_urls]
    try:
        idx = parsed_allowed_urls.index(parsed_request_url)
    except ValueError:
        idx = None
    url = allowed_urls[idx] if idx is not None else allowed_urls[0]
    # template includes tailing /, strip any trail slash here to avoid duplicates
    url = url.rstrip("/")
    return url


# Collects additional headers from flask request objects
def capture_headers(req, args_dict):
    args_dict['referer'] = req.headers.get('Referer', None)
    args_dict['origin'] = req.headers.get('Origin', None)
    args_dict['requestid'] = req.environ.get("FLASK_REQUEST_ID")
    args_dict['host'] = req.headers.get('Host', None)
    args_dict['url_root'] = req.url_root

    return args_dict

# Exceptions raised when attempting to create a
# product layer from a bad config or without correct
# product range


class ProductLayerException(Exception):
    pass


class ConfigException(Exception):
    pass

# Function wrapper for configurable functional elements


class FunctionWrapper:
    def __init__(self, product_or_style_cfg, func_cfg,
                 stand_alone=False):
        if callable(func_cfg):
            if not stand_alone:
                raise ConfigException("Directly including callable objects in configuration is no longer supported. Please reference callables by fully qualified name.")
            self._func = func_cfg
            self._args = []
            self._kwargs = {}
            self.band_mapper = None
        elif isinstance(func_cfg, str):
            self._func = get_function(func_cfg)
            self._args = []
            self._kwargs = {}
            self.band_mapper = None
        else:
            if stand_alone and callable(func_cfg["function"]):
                self._func = func_cfg["function"]
            elif callable(func_cfg["function"]):
                raise ConfigException(
                    "Directly including callable objects in configuration is no longer supported. Please reference callables by fully qualified name.")
            else:
                self._func = get_function(func_cfg["function"])
            self._args = func_cfg.get("args", [])
            self._kwargs = func_cfg.get("kwargs", {})
            if "pass_product_cfg" in func_cfg:
                print("WARNING: pass_product_cfg in function wrapper definitions has been renamed "
                      "'mapped_bands'.  Please update your config accordingly")
            if func_cfg.get("mapped_bands", func_cfg.get("pass_product_cfg", False)):
                if hasattr(product_or_style_cfg, "band_idx"):
                    # NamedLayer
                    b_idx = product_or_style_cfg.band_idx
                    self.band_mapper = b_idx.band
                else:
                    # Style
                    b_idx = product_or_style_cfg.product.band_idx
                    delocaliser = product_or_style_cfg.local_band
                    self.band_mapper = lambda b: b_idx.band(delocaliser(b))
            else:
                self.band_mapper = None

    def __call__(self, *args, **kwargs):
        if args and self._args:
            calling_args = chain(args, self._args)
        elif args:
            calling_args = args
        else:
            calling_args = self._args
        if kwargs and self._kwargs:
            calling_kwargs = self._kwargs.copy()
            calling_kwargs.update(kwargs)
        elif kwargs:
            calling_kwargs = kwargs
        else:
            calling_kwargs = self._kwargs

        if self.band_mapper:
            calling_kwargs["band_mapper"] = self.band_mapper


        return self._func(*calling_args, **calling_kwargs)


# Extent Mask Functions

def mask_by_val(data, band, val=None):
    if val is None:
        return data[band] != data[band].attrs['nodata']
    else:
        return data[band] != val


def mask_by_val2(data, band):
    # REVISIT: Is this the same as mask_by_val or subtlely different?
    return data[band] != data[band].nodata


def mask_by_bitflag(data, band):
    return ~data[band] & data[band].attrs['nodata']


def mask_by_val_in_band(data, band, mask_band, val=None):
    return mask_by_val(data, mask_band, val)


def mask_by_quality(data, band):
    return mask_by_val(data, "quality")


def mask_by_extent_flag(data, band):
    return data["extent"] == 1


def mask_by_extent_val(data, band):
    return mask_by_val(data, "extent")


def mask_by_nan(data, band):
    return ~numpy.isnan(data[band])


# Sub-product extractors

ls8_s3_path_pattern = re.compile('L8/(?P<path>[0-9]*)')


def ls8_subproduct(ds):
    return int(ls8_s3_path_pattern.search(ds.uris[0]).group("path"))

# Method for formatting urls, e.g. for use in feature_info custom inclusions.


def feature_info_url_template(data, template):
    return template.format(data=data)


def lower_get_args():
    # Get parameters in WMS are case-insensitive, and intended to be single use.
    # Spec does not specify which instance should be used if a parameter is provided more than once.
    # This function uses the LAST instance.
    d = {}
    for k in request.args.keys():
        kl = k.lower()
        for v in request.args.getlist(k):
            d[kl] = v
    return d


def create_geobox(
        crs,
        minx, miny,
        maxx, maxy,
        width=None, height=None,
):
    if width is None and height is None:
        raise Exception("Must supply at least a width or height")
    if height is not None:
        scale_y = (miny - maxy) / height
    if width is not None:
        scale_x = (maxx - minx) / width
    else:
        scale_x = -scale_y
        width = round((maxx - minx) / scale_x)
    if height is None:
        scale_y = - scale_x
        height = round((miny - maxy) / scale_y)
    affine = Affine.translation(minx, maxy) * Affine.scale(scale_x, scale_y)
    return geometry.GeoBox(width, height, affine, crs)


def xarray_image_as_png(img_data, mask=None):
    band_index = {
        "red": 1,
        "green": 2,
        "blue": 3,
        "alpha": 4,
    }
    xcoord = None
    ycoord = None
    for cc in ("x", "longitude", "Longitude", "long", "lon"):
        if cc in img_data.coords:
            xcoord = cc
            break
    for cc in ("y", "latitude", "Latitude", "lat"):
        if cc in img_data.coords:
            ycoord = cc
            break
    if not xcoord or not ycoord:
        raise Exception("Could not identify spatial coordinates")
    width = len(img_data.coords[xcoord])
    height = len(img_data.coords[ycoord])
    with MemoryFile() as memfile:
        with memfile.open(driver='PNG',
                          width=width,
                          height=height,
                          count=4,
                          transform=None,
                          dtype='uint8') as thing:
            masked = False
            last_band = None
            for band in img_data.data_vars:
                idx = band_index[band]
                band_data = img_data[band].values
                if band == "alpha" and mask is not None:
                    band_data = numpy.where(mask, band_data, 0)
                    masked = True
                elif band == "alpha":
                    masked = True
                thing.write_band(idx, band_data)
                last_band = band_data
            if not masked:
                if mask is None:
                    alpha_mask = numpy.empty(last_band.shape).astype('uint8')
                    alpha_mask.fill(255)
                else:
                    alpha_mask = numpy.where(mask, 255, 0).astype('uint8')
                thing.write_band(4, alpha_mask)
        return memfile.read()



import datetime

import pytest

import datacube_ows.ogc_utils


class DSCT:
    def __init__(self, meta):
        self.center_time = datetime.datetime(1970, 1, 1, 0, 0, 0)
        self.metadata_doc = meta


def test_dataset_center_time():
    dct = datacube_ows.ogc_utils.dataset_center_time
    ds = DSCT({})
    assert dct(ds).year == 1970
    ds = DSCT({
        "properties": {
            "dtr:start_datetime": "1980-01-01T00:00:00"
        },
    })
    assert dct(ds).year == 1980
    ds = DSCT({
        "extent": {
            "center_dt": "1990-01-01T00:00:00"
        },
        "properties": {
            "dtr:start_datetime": "1980-01-01T00:00:00"
        },
    })
    assert dct(ds).year == 1990


def test_get_service_base_url():

    # not a list
    allowed_urls = "https://foo.hello.world"
    request_url = "https://foo.bar.baz"
    ret = datacube_ows.ogc_utils.get_service_base_url(allowed_urls, request_url)
    assert ret == "https://foo.hello.world"

    # Value not in list
    allowed_urls = ["https://foo.hello.world", "https://alice.bob.eve"]
    request_url = "https://foo.bar.baz"
    ret = datacube_ows.ogc_utils.get_service_base_url(allowed_urls, request_url)
    assert ret == "https://foo.hello.world"

    # Value in list
    allowed_urls = ["https://foo.hello.world", "https://foo.bar.baz", "https://alice.bob.eve"]
    request_url = "https://foo.bar.baz"
    ret = datacube_ows.ogc_utils.get_service_base_url(allowed_urls, request_url)
    assert ret == "https://foo.bar.baz"

    # Trailing /
    allowed_urls = ["https://foo.bar.baz", "https://alice.bob.eve"]
    request_url = "https://foo.bar.baz/"
    ret = datacube_ows.ogc_utils.get_service_base_url(allowed_urls, request_url)
    assert ret == "https://foo.bar.baz"

    # include path
    allowed_urls = ["https://foo.bar.baz", "https://foo.bar.baz/wms/"]
    request_url = "https://foo.bar.baz/wms/"
    ret = datacube_ows.ogc_utils.get_service_base_url(allowed_urls, request_url)
    assert ret == "https://foo.bar.baz/wms"

    # use value from list instead of request
    allowed_urls = ["https://foo.bar.baz", "https://foo.bar.baz/wms/"]
    request_url = "http://foo.bar.baz/wms/"
    ret = datacube_ows.ogc_utils.get_service_base_url(allowed_urls, request_url)
    assert ret == "https://foo.bar.baz/wms"


def test_parse_for_base_url():
    url = "https://hello.world.bar:8000/wms/?CheckSomething"
    ret = datacube_ows.ogc_utils.parse_for_base_url(url)
    assert ret == "hello.world.bar:8000/wms"


def test_create_geobox():
    geobox = datacube_ows.ogc_utils.create_geobox("EPSG:4326",
                                                  140.7184, 145.6924, -16.1144, -13.4938,
                                                  1182, 668)
    geobox_ho = datacube_ows.ogc_utils.create_geobox("EPSG:4326",
                                                  140.7184, 145.6924, -16.1144, -13.4938,
                                                  1182, 668)
    geobox_wo = datacube_ows.ogc_utils.create_geobox("EPSG:4326",
                              140.7184, 145.6924, -16.1144, -13.4938,
                              1182, 668)
    for gb in (geobox, geobox_ho, geobox_wo):
        assert geobox.width == 1182
        assert geobox.height == 668
    with pytest.raises(Exception) as excinfo:
        geobox_no = datacube_ows.ogc_utils.create_geobox("EPSG:4326",
                                                         140.7184, 145.6924, -16.1144, -13.4938)
    assert "Must supply at least a width or height" in str(excinfo.value)


from tests.utils import dummy_da

coords = [
    ("x", [-1.0, -0.5, 0.0, 0.5, 1.0]),
]


def test_mask_by_val():
    data = {
        "match": dummy_da(-999, "match", coords, attrs={"nodata": -999}, dtype="int16"),
        "dont_match": dummy_da(679, "dont_match", coords, attrs={"nodata": -999}, dtype="int16"),
    }
    mask = datacube_ows.ogc_utils.mask_by_val(data, "match")
    assert not mask.values[0]
    mask = datacube_ows.ogc_utils.mask_by_val(data, "dont_match")
    assert mask.values[0]
    mask = datacube_ows.ogc_utils.mask_by_val(data, "match", val=679)
    assert mask.values[0]
    mask = datacube_ows.ogc_utils.mask_by_val(data, "dont_match", val=679)
    assert not mask.values[0]


def test_mask_by_val2():
    data = {
        "match": dummy_da(-999, "match", coords, attrs={"nodata": -999}, dtype="int16"),
        "dont_match": dummy_da(679, "dont_match", coords, attrs={"nodata": -999}, dtype="int16"),
    }
    mask = datacube_ows.ogc_utils.mask_by_val(data, "match")
    assert not mask.values[0]
    mask = datacube_ows.ogc_utils.mask_by_val(data, "dont_match")
    assert mask.values[0]


def test_mask_by_bitflag():
    data = {
        "match": dummy_da(128, "match", coords, attrs={"nodata": 128}, dtype="uint8"),
        "dont_match": dummy_da(63, "dont_match", coords, attrs={"nodata": 128}, dtype="uint8"),
    }
    mask = datacube_ows.ogc_utils.mask_by_bitflag(data, "match")
    assert not mask.values[0]
    mask = datacube_ows.ogc_utils.mask_by_bitflag(data, "dont_match")
    assert mask.values[0]


def test_mask_by_val_in_band():
    data = {
        "match": dummy_da(-999, "match", coords, attrs={"nodata": -999}, dtype="int16"),
        "dont_match": dummy_da(679, "dont_match", coords, attrs={"nodata": -999}, dtype="int16"),
        "dband": dummy_da(0.77, "dband", coords, dtype="float128"),
    }
    mask = datacube_ows.ogc_utils.mask_by_val_in_band(data, "dband", mask_band="match")
    assert not mask.values[0]
    mask = datacube_ows.ogc_utils.mask_by_val_in_band(data, "dband", mask_band="dont_match", val=679)
    assert not mask.values[0]


def test_mask_by_quality():
    data = {
        "quality": dummy_da(-999, "match", coords, attrs={"nodata": -999}, dtype="int16"),
        "dband": dummy_da(0.77, "dband", coords, dtype="float128"),
    }
    mask = datacube_ows.ogc_utils.mask_by_quality(data, "dband")
    assert not mask.values[0]


def mask_by_extent_flag(data, band):
    data = {
        "extent": dummy_da(1, "match", coords, dtype="uint8"),
        "dband": dummy_da(0.77, "dband", coords, dtype="float128"),
    }
    mask = datacube_ows.ogc_utils.mask_by_extent_flag(data, "dband")
    assert mask.values[0]
    data["extent"] = dummy_da(0, "match", coords, dtype="uint8"),
    mask = datacube_ows.ogc_utils.mask_by_extent_flag(data, "dband")
    assert not mask.values[0]


def test_mask_by_extent_val():
    data = {
        "extent": dummy_da(-999, "match", coords, attrs={"nodata": -999}, dtype="int16"),
        "dband": dummy_da(0.77, "dband", coords, dtype="float128"),
    }
    mask = datacube_ows.ogc_utils.mask_by_extent_val(data, "dband")
    assert not mask.values[0]


def test_mask_by_nan():
    data = {
        "match": dummy_da(float("nan"), "match", coords, dtype="float128"),
        "dont_match": dummy_da(67.9, "dont_match", coords, dtype="float128"),
    }
    mask = datacube_ows.ogc_utils.mask_by_nan(data, "match")
    assert not mask.values[0]
    mask = datacube_ows.ogc_utils.mask_by_nan(data, "dont_match")
    assert mask.values[0]



from unittest.mock import MagicMock

import pytest

from datacube_ows.ogc_utils import ConfigException
from datacube_ows.tile_matrix_sets import TileMatrixSet


@pytest.fixture
def wwwm_tms_cfg():
    return TileMatrixSet.default_tm_sets["WholeWorld_WebMercator"].copy()


@pytest.fixture
def tmsmin_global_cfg():
    gcfg = MagicMock()
    gcfg.published_CRSs = {
            "EPSG:3857": {  # Web Mercator
                "geographic": False,
                "horizontal_coord": "x",
                "vertical_coord": "y",
                "vertical_coord_first": False
            },
            "I:CANT:BELIEVE:ITS:NOT:EPSG:3857": {  # I Can't Believe It's Not Web Mercator
                "geographic": False,
                "horizontal_coord": "x",
                "vertical_coord": "y",
                "vertical_coord_first": True
            },
            "EPSG:4326": {  # WGS-84
                "geographic": True,
                "vertical_coord_first": True
            },
            "EPSG:3111": {  # VicGrid94 for delwp.vic.gov.au
                "geographic": False,
                "horizontal_coord": "x",
                "vertical_coord": "y",
                "vertical_coord_first": False
            },
    }
    return gcfg


def test_tms_unpublished_crs(wwwm_tms_cfg):
    global_cfg = MagicMock()
    global_cfg.published_CRSs = {}
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "unpublished CRS" in str(excinfo.value)
    assert "EPSG:3857" in str(excinfo.value)


def test_matrix_origin_float_array(wwwm_tms_cfg, tmsmin_global_cfg):
    global_cfg = tmsmin_global_cfg
    wwwm_tms_cfg["matrix_origin"] = None
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "Matrix origin" in str(excinfo.value)
    assert "must be a list" in str(excinfo.value)

    wwwm_tms_cfg["matrix_origin"] = [3.3, 6.3, 7.8]
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "Matrix origin" in str(excinfo.value)
    assert "must have two values" in str(excinfo.value)

    wwwm_tms_cfg["matrix_origin"] = [6.3, "spaghetti bolognaise"]
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "Matrix origin" in str(excinfo.value)
    assert "spaghetti bolognaise" in str(excinfo.value)
    assert "non-float" in str(excinfo.value)
    assert "str" in str(excinfo.value)


def test_tile_size_int_array(wwwm_tms_cfg, tmsmin_global_cfg):
    global_cfg = tmsmin_global_cfg
    wwwm_tms_cfg["tile_size"] = None
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "Tile size" in str(excinfo.value)
    assert "must be a list" in str(excinfo.value)

    wwwm_tms_cfg["tile_size"] = [256, 512, 1024]
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "Tile size" in str(excinfo.value)
    assert "must have two values" in str(excinfo.value)

    wwwm_tms_cfg["tile_size"] = [64, "spaghetti bolognaise"]
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "Tile size" in str(excinfo.value)
    assert "spaghetti bolognaise" in str(excinfo.value)
    assert "non-int" in str(excinfo.value)
    assert "str" in str(excinfo.value)


def test_scale_set_array(wwwm_tms_cfg, tmsmin_global_cfg):
    global_cfg = tmsmin_global_cfg
    wwwm_tms_cfg["scale_set"] = None
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "scale_set" in str(excinfo.value)
    assert "is not a list" in str(excinfo.value)
    wwwm_tms_cfg["scale_set"] = []
    with pytest.raises(ConfigException) as excinfo:
        tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert "test" in str(excinfo.value)
    assert "scale_set" in str(excinfo.value)
    assert "no scale denominators" in str(excinfo.value)


def test_tms_crs_cfg(wwwm_tms_cfg, tmsmin_global_cfg):
    global_cfg = tmsmin_global_cfg
    tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert not tms.crs_cfg["vertical_coord_first"]


def test_tms_crs_display(wwwm_tms_cfg, tmsmin_global_cfg):
    global_cfg = tmsmin_global_cfg
    tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert tms.crs_display == "urn:ogc:def:crs:EPSG::3857"
    wwwm_tms_cfg["crs"] = "I:CANT:BELIEVE:ITS:NOT:EPSG:3857"
    tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert tms.crs_display == "I:CANT:BELIEVE:ITS:NOT:EPSG:3857"
    wwwm_tms_cfg["crs"] = "EPSG:3857"
    wwwm_tms_cfg["force_raw_crs_name"] = True
    tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert tms.crs_display == "EPSG:3857"


def test_tms_exponent(wwwm_tms_cfg, tmsmin_global_cfg):
    global_cfg = tmsmin_global_cfg
    tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert tms.width_exponent(0) == 0
    assert tms.width_exponent(1) == 1
    assert tms.width_exponent(12) == 12
    assert tms.height_exponent(0) == 0
    assert tms.height_exponent(1) == 1
    assert tms.height_exponent(12) == 12
    wwwm_tms_cfg["matrix_exponent_initial_offsets"] = [-2, 2]
    tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    assert tms.width_exponent(0) == 0
    assert tms.width_exponent(1) == 0
    assert tms.width_exponent(2) == 0
    assert tms.width_exponent(3) == 1
    assert tms.width_exponent(12) == 10
    assert tms.height_exponent(0) == 2
    assert tms.height_exponent(1) == 3
    assert tms.height_exponent(12) == 14


def test_tms_wms_bbox(wwwm_tms_cfg, tmsmin_global_cfg):
    global_cfg = tmsmin_global_cfg
    tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    a, b, c, d = tms.wms_bbox_coords(7, 32, 24)
    assert a == pytest.approx(-12523442.7142, 0.001)
    assert b == pytest.approx(9705668.103538, 0.001)
    assert c == pytest.approx(-12210356.64638, 0.001)
    assert d == pytest.approx(10018754.17139, 0.001)
    wwwm_tms_cfg["crs"] = "I:CANT:BELIEVE:ITS:NOT:EPSG:3857"
    tms = TileMatrixSet("test", wwwm_tms_cfg, global_cfg)
    a, b, c, d = tms.wms_bbox_coords(7, 32, 24)
    assert b == pytest.approx(-12523442.7142, 0.001)
    assert a == pytest.approx(9705668.103538, 0.001)
    assert d == pytest.approx(-12210356.64638, 0.001)
    assert c == pytest.approx(10018754.17139, 0.001)



import numpy as np
import pandas as pd
import pytest
import xarray as xr

from datacube_ows.ogc_utils import ConfigException
from datacube_ows.styles.base import StyleDefBase


def test_multidate_handler():
    # TODO: Consolidate these into a fixture
    class FakeData:
        def __init__(self):
            self.nodata = np.nan

        def item(self):
            return np.nan

    class FakeDataset:
        def __getitem__(self, key):
            return FakeData()

    class FakeMdhStyle:
        include_in_feature_info = True

        def __init__(self):
            self.product = "test"
            self.needed_bands = ["test"]
            self.index_function = lambda x: FakeData()

    data = np.random.randint(0, 255, size=(4, 3), dtype=np.uint8)
    locs = ["IA", "IL", "IN"]
    times = pd.date_range("2000-01-01", periods=4)
    fake_mask = xr.DataArray(data, coords=[times, locs], dims=["time", "space"])

    fake_cfg = {
        "allowed_count_range": [0, 10],
        "aggregator_function": "datacube_ows.band_utils.multi_date_delta",
    }

    fake_cfg_equal = {
        "allowed_count_range": [1, 1],
        "aggregator_function": "datacube_ows.band_utils.multi_date_delta",
    }

    mdh = StyleDefBase.MultiDateHandler(FakeMdhStyle(), fake_cfg)
    assert mdh is not None
    assert not mdh.legend(None)
    assert mdh.collapse_mask(fake_mask) is not None
    assert isinstance(mdh.range_str(), str)
    assert mdh.applies_to(2)
    assert not mdh.applies_to(11)

    mdh_equal = StyleDefBase.MultiDateHandler(FakeMdhStyle(), fake_cfg_equal)
    assert isinstance(mdh_equal.range_str(), str)

    with pytest.raises(ConfigException) as excinfo:
        bad_mdh = StyleDefBase.MultiDateHandler(FakeMdhStyle(), {})

    assert "must have an allowed_count_range" in str(excinfo.value)

    with pytest.raises(ConfigException) as excinfo:
        bad_mdh = StyleDefBase.MultiDateHandler(
            FakeMdhStyle(), {"allowed_count_range": [0, 5, 10], }
        )

    assert "allowed_count_range must have 2" in str(excinfo.value)

    with pytest.raises(ConfigException) as excinfo:
        bad_mdh = StyleDefBase.MultiDateHandler(
            FakeMdhStyle(), {"allowed_count_range": [10, 5], }
        )

    assert "minimum must be less than equal to maximum" in str(excinfo.value)

    with pytest.raises(ConfigException) as excinfo:
        bad_mdh = StyleDefBase.MultiDateHandler(
            FakeMdhStyle(), {"allowed_count_range": [0, 10], }
        )

    assert "Aggregator function is required" in str(excinfo.value)

    with pytest.raises(NotImplementedError) as excinfo:
        mdh.transform_data(None)

    assert str(excinfo) == "<ExceptionInfo NotImplementedError() tblen=2>"



from datetime import datetime

import pytest
import pytz

from datacube_ows.ows_configuration import (TIMERES_MON, TIMERES_RAW,
                                            TIMERES_YR, OWSProductLayer)


def dummy_timeres_layer(time_res):
    prod = product_layer = OWSProductLayer.__new__(OWSProductLayer)
    prod.time_resolution = time_res
    return prod


class Thing:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


@pytest.fixture
def dummy_raw_layer():
    return dummy_timeres_layer(TIMERES_RAW)


@pytest.fixture
def dummy_monthly_layer():
    return dummy_timeres_layer(TIMERES_MON)


@pytest.fixture
def dummy_yearly_layer():
    return dummy_timeres_layer(TIMERES_YR)


@pytest.fixture
def simple_geobox():
    from affine import Affine
    from datacube.utils import geometry

    aff = Affine.translation(145.0, -35.0) * Affine.scale(
        1.0 / 256, 2.0 / 256
    )
    return geometry.GeoBox(256, 256, aff, 'EPSG:4326')


def test_raw_timeres(dummy_raw_layer, simple_geobox):
    assert dummy_raw_layer.is_raw_time_res
    assert not dummy_raw_layer.is_month_time_res
    assert not dummy_raw_layer.is_year_time_res
    assert dummy_raw_layer.dataset_groupby() == "solar_day"

    assert dummy_raw_layer.search_times(
        datetime(2020, 6, 7, 20, 20, 0, tzinfo=pytz.utc),
        simple_geobox,
    ) == (
        datetime(2020, 6, 6, 13, 55, tzinfo=pytz.utc),
        datetime(2020, 6, 7, 13, 54, 59, tzinfo=pytz.utc),
    )


def test_mon_timeres(dummy_monthly_layer, simple_geobox):
    assert not dummy_monthly_layer.is_raw_time_res
    assert dummy_monthly_layer.is_month_time_res
    assert not dummy_monthly_layer.is_year_time_res
    gby = dummy_monthly_layer.dataset_groupby()
    assert gby.dimension == 'time'

    t = Thing(begin="ABC")
    ds = Thing(time=t)
    assert gby.group_by_func(ds) == "ABC"
    assert gby.units == 'seconds since 1970-01-01 00:00:00'
    assert gby.sort_key(ds) == "ABC"

    assert dummy_monthly_layer.search_times(
        datetime(2020, 6, 1, tzinfo=pytz.utc),
        simple_geobox,
    ) == (
               datetime(2020, 6, 1),
               datetime(2020, 6, 30),
           )


def test_year_timeres(dummy_yearly_layer):
    assert not dummy_yearly_layer.is_raw_time_res
    assert not dummy_yearly_layer.is_month_time_res
    assert dummy_yearly_layer.is_year_time_res
    gby = dummy_yearly_layer.dataset_groupby()
    assert gby.dimension == 'time'

    t = Thing(begin="ABC")
    ds = Thing(time=t)
    assert gby.group_by_func(ds) == "ABC"
    assert gby.units == 'seconds since 1970-01-01 00:00:00'
    assert gby.sort_key(ds) == "ABC"

    assert dummy_yearly_layer.search_times(
        datetime(2020, 6, 1, tzinfo=pytz.utc),
        simple_geobox,
    ) == (
               datetime(2020, 1, 1),
               datetime(2020, 12, 31, 23, 59, 59),
           )



"""
Test creation of colour maps from matplotlib
"""
from datacube_ows.ows_cfg_example import style_deform
from datacube_ows.styles.ramp import read_mpl_ramp


def test_get_mpl_cmap():
    matplotlib_ramp_name = style_deform['mpl_ramp']
    assert matplotlib_ramp_name
    ows_ramp_dict = read_mpl_ramp(matplotlib_ramp_name)
    assert len(ows_ramp_dict) == 11
    for cmap in ows_ramp_dict:
        assert "color" in cmap
        assert "value" in cmap
        assert cmap["color"].startswith("#")
        assert isinstance(cmap["value"], float)



from unittest.mock import MagicMock

import pytest

from datacube_ows.ogc_exceptions import WCS2Exception
from datacube_ows.wcs2_utils import uniform_crs


@pytest.fixture
def minimal_cfg():
    cfg = MagicMock()
    cfg.published_CRSs = {
        "dummy": {},
    }
    return cfg


def test_uniform_crs_url(minimal_cfg):
    crs = uniform_crs(minimal_cfg, "http://www.opengis.net/def/crs/EPSG/666")
    assert crs == "EPSG:666"


def test_uniform_crs_urn(minimal_cfg):
    crs = uniform_crs(minimal_cfg, "urn:ogc:def:crs:EPSG:666")
    assert crs == "EPSG:666"


def test_uniform_crs_epsg(minimal_cfg):
    crs = uniform_crs(minimal_cfg, "EPSG:666")
    assert crs == "EPSG:666"


def test_uniform_crs_published(minimal_cfg):
    crs = uniform_crs(minimal_cfg, "dummy")
    assert crs == "dummy"


def test_uniform_crs_published_with_exception(minimal_cfg):
    with pytest.raises(WCS2Exception) as e:
        crs = uniform_crs(minimal_cfg, "spam")
    assert "spam" in str(e.value)
    assert "Not a CRS" in str(e.value)



import os

import pytest

from datacube_ows.cfg_parser import main


def test_cfg_parser_simple(runner):
    result = runner.invoke(main, [])
    assert result.exit_code == 0


def test_cfg_parser_parse_only(runner):
    result = runner.invoke(main, ["-p"])
    assert result.exit_code == 0


def test_cfg_parser_folder_hierarchy(runner):
    result = runner.invoke(main, ["-f"])
    assert result.exit_code == 0


def test_cfg_parser_styles(runner):
    result = runner.invoke(main, ["-s"])
    assert result.exit_code == 0


def test_cfg_parser_folder_hierarchy_and_styles(runner):
    result = runner.invoke(main, ["-f", "-s"])
    assert result.exit_code == 0


def test_cfg_parser_folders_parse_only(runner):
    result = runner.invoke(main, ["-f", "-p"])
    assert result.exit_code == 1


def test_cfg_parser_input_file_compare(runner):
    this_dir = os.path.dirname(os.path.dirname(__file__))
    result = runner.invoke(main, ["-i", f"{this_dir}/ows_cfg_report.json"])
    assert result.exit_code == 0


@pytest.mark.xfail(reason="Permission denied")
def test_cfg_parser_output_file_compare(runner):
    result = runner.invoke(main, ["-o", "inventory.json"])
    assert result.exit_code == 0


def test_cfg_parser_version(runner):
    result = runner.invoke(main, ["--version"])
    assert result.exit_code == 0



"""Test update ranges on DB using Click testing
https://click.palletsprojects.com/en/7.x/testing/
"""
from datacube_ows.update_ranges import main


def test_updates_ranges_schema(runner, role_name):
    result = runner.invoke(main, ["--schema", "--role", role_name])
    assert "Cannot find SQL resource" not in result.output
    assert result.exit_code == 0


def test_update_ranges_views(runner):
    result = runner.invoke(main, ["--views", "--blocking"])
    assert "Cannot find SQL resource" not in result.output
    assert result.exit_code == 0

    result = runner.invoke(main, ["--views"])
    assert "Cannot find SQL resource" not in result.output
    assert result.exit_code == 0


def test_update_version(runner):
    result = runner.invoke(main, ["--version"])
    assert "Open Data Cube Open Web Services (datacube-ows) version" in result.output
    assert result.exit_code == 0


def test_update_ranges_product(runner, product_name):
    result = runner.invoke(main, [product_name])
    assert "ERROR" not in result.output
    assert result.exit_code == 0


def test_update_ranges(runner):
    result = runner.invoke(main)
    assert "ERROR" not in result.output
    assert result.exit_code == 0


def test_update_ranges_misuse_cases(runner, role_name, product_name):
    result = runner.invoke(main, ["--schema"])
    assert "Sorry" in result.output
    assert result.exit_code == 0

    result = runner.invoke(main, ["--role", role_name])
    assert "Sorry" in result.output
    assert result.exit_code == 0

    result = runner.invoke(main, ["--views", product_name])
    assert "Sorry" in result.output
    assert result.exit_code == 0

    result = runner.invoke(main, ["--schema", product_name])
    assert "Sorry" in result.output
    assert result.exit_code == 0





bands_c3_ls_common = {
    "nbart_blue": ["nbart_blue"],
    "nbart_green": ["nbart_green"],
    "nbart_red": ["nbart_red"],
    "nbart_nir": ["nbart_nir", "nbart_near_infrared"],
    "nbart_swir_1": ["nbart_swir_1", "nbart_shortwave_infrared_1"],
    "nbart_swir_2": ["nbart_swir_2", "nbart_shortwave_infrared_2"],
    "oa_nbart_contiguity": ["oa_nbart_contiguity", "nbart_contiguity"],
    "oa_fmask": ["oa_fmask", "fmask"],
}

bands_c3_ls_7 = bands_c3_ls_common.copy()
bands_c3_ls_7.update({
    "nbart_panchromatic": [],
})


bands_c3_ls_8 = bands_c3_ls_7.copy()
bands_c3_ls_8.update({
    "nbart_coastal_aerosol": ["coastal_aerosol",  "nbart_coastal_aerosol"],
})





from pyproj import CRS

SUPPORTED_CRS = [
   'EPSG:3857', # Web Mercator
   'EPSG:4326', # WGS-84
   'EPSG:3577', # GDA-94
   'EPSG:3111', # VicGrid94
   'EPSG:32648', # WGS 84 / Cambodiacube
   'ESRI:102022', # Africa
   # 'EPSG:102022', # Depreciated Africa
   'EPSG:6933', # Africa
]


def test_pyproj_crs():
   for crs_string in SUPPORTED_CRS:
      try:
         crs = CRS(crs_string)
         assert crs is not None
      except:
         assert False




import os

pytest_plugins = ["helpers_namespace"]
import pytest
from click.testing import CliRunner
from pytest_localserver.http import WSGIServer

from datacube_ows import ogc
from datacube_ows.ogc import app


@pytest.fixture
def flask_client():
    with app.test_client() as client:
        yield client


class generic_obj(object):
    pass


@pytest.fixture(scope="session")
def ows_server(request):
    """
    Run the OWS server for the duration of these tests
    """
    external_url = os.environ.get("SERVER_URL")
    if external_url:
        server = generic_obj()
        server.url = external_url
    else:
        server = WSGIServer(port="5000", application=ogc.app)
        server.start()
        request.addfinalizer(server.stop)

    return server


@pytest.fixture
def runner():
    return CliRunner()


@pytest.helpers.register
def enclosed_bbox(bbox):
    lon_min, lat_min, lon_max, lat_max = bbox
    lon_range = lon_max - lon_min
    lat_range = lat_max - lat_min

    return (
        lon_min + 0.45 * lon_range,
        lat_min + 0.45 * lat_range,
        lon_max - 0.45 * lon_range,
        lat_max - 0.45 * lat_range,
    )


@pytest.helpers.register
def disjoint_bbox(bbox):
    lon_min, lat_min, lon_max, lat_max = bbox
    lon_range = lon_max - lon_min
    lat_range = lat_max - lat_min

    return (
        lon_min - 0.4 * lon_range,
        lat_min - 0.4 * lat_range,
        lon_min - 0.2 * lon_range,
        lat_min - 0.2 * lat_range,
    )


@pytest.fixture
def product_name():
    return "ls8_usgs_level1_scene_layer"


@pytest.fixture
def role_name():
    return os.getenv("DB_USERNAME")


@pytest.fixture
def multiproduct_name():
    return "fc_albers_combined"



from __future__ import absolute_import, division, print_function

import logging
from contextlib import contextmanager
from threading import Lock

from datacube import Datacube

_LOG = logging.getLogger(__name__)


class CubePool():
    _instances = {}
    _cubes = None
    _cubes_lock = None

    def __new__(cls, app):
        if app not in cls._instances:
            cls._instances[app] = super(CubePool, cls).__new__(cls)
        return cls._instances[app]

    def __init__(self, app):
        self.app = app
        if not self._cubes:
            self._cubes = {}
        if not self._cubes_lock:
            self._cubes_lock = Lock()

    def get_cube(self):
        self._cubes_lock.acquire()
        for c, assigned in self._cubes.items():
            if not assigned:
                self._cubes[c] = True
                self._cubes_lock.release()
                return c
        try:
            c = self._new_cube()
            self._cubes[c] = True
        # pylint: disable=broad-except
        except Exception as e:
            _LOG.error("ODC initialisation failed: %s", str(e))
            c = None
        finally:
            self._cubes_lock.release()
        return c

    def release_cube(self, c):
        if c not in self._cubes:
            raise Exception("Cannot release non-pool datacube.")
        self._cubes_lock.acquire()
        self._cubes[c] = False
        self._cubes_lock.release()

    def _new_cube(self):
        return Datacube(app=self.app)

    def __len__(self):
        return len(self._cubes)


def get_cube(app="wms"):
    pool = CubePool(app=app)
    return pool.get_cube()


def release_cube(c, app="wms"):
    pool = CubePool(app=app)
    return pool.release_cube(c)


def pool_size(app="wms"):
    pool = CubePool(app=app)
    return len(pool)


@contextmanager
def cube(app="wms"):
    dc = get_cube(app)
    try:
        yield dc
    finally:
        if dc:
            release_cube(dc, app)



from urllib import request

import pytest
import requests
from lxml import etree
from owslib.wms import WebMapService


def get_xsd(name):
    xsd_f = open("wms_xsds/" + name)
    schema_doc = etree.parse(xsd_f)

    return etree.XMLSchema(schema_doc)


def check_wms_error(url, expected_error_message=None, expected_status_code=400):
    try:
        resp = request.urlopen(url, timeout=10)

        # Should not get here
        assert False
    except Exception as e:
        # Validate status code
        assert e.getcode() == expected_status_code

        resp_content = e.fp.read()
        resp_xml = etree.XML(resp_content)

        # Validate response against Schema
        ex_xsd = get_xsd("exceptions_1_3_0.xsd")
        assert ex_xsd.validate(resp_xml)

        # Confirm error message is appropriate, ignore case
        if expected_error_message:
            assert (
                resp_xml[0].text.strip().casefold() == expected_error_message.casefold()
            )


def test_no_request(ows_server):
    # Make empty request to server:
    check_wms_error(ows_server.url + "/wms", "No operation specified", 400)


def test_invalid_operation(ows_server):
    # Make invalid operation request to server:
    check_wms_error(
        ows_server.url + "/wms?request=NoSuchOperation",
        "Unrecognised operation: NoSuchOperation",
        400,
    )


def test_getcap_badsvc(ows_server):
    # Make bad service request to server:
    check_wms_error(
        ows_server.url + "/wms?request=GetCapabilities&service=NotWMS",
        "Invalid service",
        400,
    )


def test_getcap(ows_server):
    resp = request.urlopen(
        ows_server.url + "/wms?request=GetCapabilities&service=WMS&version=1.3.0",
        timeout=10,
    )

    # Confirm success
    assert resp.code == 200

    # Validate XML Schema
    resp_xml = etree.parse(resp.fp)
    gc_xds = get_xsd("capabilities_1_3_0.xsd")
    assert gc_xds.validate(resp_xml)


def test_getcap_coord_order(ows_server):
    resp = request.urlopen(
        ows_server.url + "/wms?request=GetCapabilities&service=WMS&version=1.3.0",
        timeout=10,
    )

    # Confirm success
    assert resp.code == 200

    # Validate XML Schema
    resp_xml = etree.parse(resp.fp)
    root = resp_xml.getroot()
    layers = root.findall(".//{http://www.opengis.net/wms}Layer[@queryable='1']")
    for layer in layers:
        wLong = layer.findall(".//{http://www.opengis.net/wms}westBoundLongitude")[0]
        geo_bbox = layer.findall(
            "./{http://www.opengis.net/wms}BoundingBox[@CRS='EPSG:4326']"
        )[0]
        assert wLong.text == geo_bbox.attrib["miny"]


def test_wms_server(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WMS service
    wms = WebMapService(url=ows_server.url + "/wms", version="1.3.0")

    assert wms.identification.type == "WMS"

    # Ensure that we have at least some layers available
    contents = list(wms.contents)
    assert contents


def test_wms_getmap(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WMS service
    wms = WebMapService(url=ows_server.url + "/wms", version="1.3.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wms.contents)
    test_layer_name = contents[0]
    test_layer = wms.contents[test_layer_name]

    bbox = test_layer.boundingBoxWGS84

    img = wms.getmap(
        layers=[test_layer_name],
        styles=[],
        srs="EPSG:4326",
        bbox=pytest.helpers.enclosed_bbox(bbox),
        size=(150, 150),
        format="image/png",
        transparent=True,
        time=test_layer.timepositions[len(test_layer.timepositions) // 2].strip(),
    )
    assert img
    assert img.info()["Content-Type"] == "image/png"

    img = wms.getmap(
        layers=[test_layer_name],
        styles=[],
        srs="I-CANT-BELIEVE-ITS-NOT-EPSG:4326",
        bbox=pytest.helpers.enclosed_bbox(bbox),
        size=(150, 150),
        format="image/png",
        transparent=True,
        time=test_layer.timepositions[len(test_layer.timepositions) // 2].strip(),
    )
    assert img
    assert img.info()["Content-Type"] == "image/png"


def test_wms_multiproduct_getmap(ows_server, multiproduct_name):
    # Use owslib to confirm that we have a somewhat compliant WMS service
    wms = WebMapService(url=ows_server.url + "/wms", version="1.3.0", timeout=120)

    # Run test against dedicated multiproduct
    test_layer = wms.contents[multiproduct_name]

    bbox = test_layer.boundingBoxWGS84

    img = wms.getmap(
        layers=[multiproduct_name],
        styles=[],
        srs="EPSG:3577",
        bbox=pytest.helpers.enclosed_bbox(bbox),
        size=(150, 150),
        format="image/png",
        transparent=True,
        time=test_layer.timepositions[len(test_layer.timepositions) // 2].strip(),
    )
    assert img
    assert img.info()["Content-Type"] == "image/png"


def test_wms_multidate_getmap(ows_server):
    # This one will only work with specially prepared test data, sorry.
    wms = WebMapService(url=ows_server.url + "/wms", version="1.3.0", timeout=120)

    img = wms.getmap(
        layers=["ls8_usgs_level1_scene_layer"],
        styles=["ndvi_delta"],
        srs="EPSG:4326",
        bbox=(145.75, -44.2,
              148.69, -42.11),
        size=(150, 150),
        format="image/png",
        transparent=True,
        time="2019-01-30,2019-03-03",
    )
    assert img.info()["Content-Type"] == "image/png"


def test_wms_style_looping_getmap(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WMS service
    wms = WebMapService(url=ows_server.url + "/wms", version="1.3.0", timeout=120)

    # Ensure that we have at least some layers available
    contents = list(wms.contents)
    test_layer_name = contents[0]
    test_layer = wms.contents[test_layer_name]

    test_layer_styles = wms.contents[test_layer_name].styles

    bbox = test_layer.boundingBoxWGS84
    layer_bbox = pytest.helpers.enclosed_bbox(bbox)
    layer_time = test_layer.timepositions[len(test_layer.timepositions) // 2].strip()

    for style in test_layer_styles:
        img = wms.getmap(
            layers=[test_layer_name],
            styles=[style],
            srs="EPSG:4326",
            bbox=layer_bbox,
            size=(150, 150),
            format="image/png",
            transparent=True,
            time=layer_time,
        )
        assert img.info()["Content-Type"] == "image/png"


def test_wms_getfeatureinfo(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WMS service
    wms = WebMapService(url=ows_server.url + "/wms", version="1.3.0")

    # Ensure that we have at least some layers available
    contents = list(wms.contents)
    test_layer_name = contents[0]
    test_layer = wms.contents[test_layer_name]

    bbox = test_layer.boundingBoxWGS84
    response = wms.getfeatureinfo(
        layers=[test_layer_name],
        srs="EPSG:4326",
        bbox=pytest.helpers.enclosed_bbox(bbox),
        size=(256, 256),
        format="image/png",
        query_layers=[test_layer_name],
        info_format="application/json",
        xy=(250, 250),
    )

    assert response
    assert response.info()["Content-Type"] == "application/json"


def test_wms_getlegend(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WMS service
    wms = WebMapService(url=ows_server.url + "/wms", version="1.3.0")

    # Ensure that we have at least some layers available
    contents = list(wms.contents)
    test_layer_name = contents[0]

    test_layer_styles = wms.contents[test_layer_name].styles
    for style in test_layer_styles:
        # check if this layer has a legend
        legend_url = test_layer_styles[style].get("legend")
        if legend_url:
            resp = requests.head(legend_url, allow_redirects=False)
            assert resp.headers.get("content-type") == "image/png"


def test_wms_getlegendgraphic(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WMS service
    wms = WebMapService(url=ows_server.url + "/wms", version="1.3.0")

    # Ensure that we have at least some layers available
    contents = list(wms.contents)
    test_layer_name = contents[0]

    test_layer_styles = wms.contents[test_layer_name].styles
    for style in test_layer_styles:
        # check if this layer has a legend
        legend_url = test_layer_styles[style].get("legend")
        url = ows_server.url + "/wms"
        resp = requests.get(
            url,
            allow_redirects=False,
            params={
                "request": "GetLegendGraphic",
                "layer": test_layer_name,
                "version": "1.3.0",
                "service": "WMS",
                "styles": style,
                "format": "image/png",
            },
        )
        if legend_url:
            assert resp.headers.get("content-type") == "image/png"
            assert resp.status_code == 200
        else:
            assert resp.status_code == 404



from datacube_ows.config_toolkit import deepinherit


def test_deepinherit_shallow():
    parent = {
        "a": 72,
        "b": "eagle",
        "c": False
    }

    child = {
        "a": 365
    }
    child = deepinherit(parent, child)
    assert child['a'] == 365
    assert child["b"] == "eagle"
    assert not child["c"]


def test_deepinherit_deep():
    parent = {
        "a": 72,
        "b": {
            "fruit": "grapes",
            "spice": "cummin",
            "cake": "chocolate",
            "y": ["some", "body", "once"],
            "z": [44, 42, 53],
            "c": {
                "foo": "bar",
                "wing": "wang"
            }
        }
    }

    child = {
        "b": {
            "spice": "nutmeg",
            "c": {
                "wing": "chicken"
            },
            "y": ["told", "me"],
            "z": [11]
        }
    }
    child = deepinherit(parent, child)
    assert child["a"] == 72
    assert child["b"]["spice"] == "nutmeg"
    assert child["b"]["fruit"] == "grapes"
    assert child["b"]["c"]["foo"] == "bar"
    assert child["b"]["c"]["wing"] == "chicken"
    assert child["b"]["z"] == [11]
    assert child["b"]["y"] == ["some", "body", "once", "told", "me"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

requirements = [
    'datacube',
    'Flask',
    'flask_log_request_id',
    'requests',
    'affine',
    'click',
    'colour',
    'lxml',
    'matplotlib',
    'numpy',
    'Pillow',
    'prometheus_client',
    'psycopg2',
    'python_dateutil',
    'pytz',
    'rasterio',
    'regex',
    'scikit-image',
    'timezonefinderL',
    'python-slugify',
    'geoalchemy',
    'xarray',
    'pyows',
    'prometheus-flask-exporter',
    #
    'setuptools_scm'
]

test_requirements = [
    # TODO: put package test requirements here
    'pytest', 'pytest-cov', 'pytest_localserver', 'owslib', 'mock', 'pep8', 'pylint==1.6.4', 'pytest-helpers-namespace'
]

setup(
    name='datacube_ows',
    description="Open Data Cube Open Web Services",
    long_description="""
============
datacube-ows
============

Open Web Services for the Open Datacube.

* Free software: Apache Software License 2.0
* Documentation: https://datacube-ows.readthedocs.io.

Features
--------

* Leverages the power of the Open Data Cube, including support for COGs on S3.
* Supports WMS and WMTS.
* Experimental support for WCS (1.0, 2.0, 2.1).
    
    """,
    author="Open Data Cube",
    author_email='earth.observation@ga.gov.au',
    url='https://github.com/opendatacube/datacube-ows',
    entry_points={
        'console_scripts': [
            'datacube-ows=datacube_ows.wsgi:main',
            'datacube-ows-update=datacube_ows.update_ranges:main',
            'datacube-ows-cfg-parse=datacube_ows.cfg_parser:main'
        ]
    },
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='datacube, wms, wcs',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
    ],
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
    test_suite='tests',
    tests_require=test_requirements
)



#pylint: skip-file
import os
import sys

# This is the directory of the source code that the web app will run from
sys.path.append("/opt")

# The location of the datcube config file.
os.environ.setdefault("DATACUBE_CONFIG_PATH", "/opt/odc/.datacube.conf.local")

from datacube_ows.ogc import app

application = app


def main():
    app.run()


if __name__ == '__main__':
    main()



import json
from enum import Enum

from datacube.utils.geometry import Geometry as ODCGeom
from geoalchemy2 import Geometry
from psycopg2.extras import DateTimeTZRange
from sqlalchemy import SMALLINT, Column, MetaData, Table, or_, select, text
from sqlalchemy.dialects.postgresql import TSTZRANGE, UUID
from sqlalchemy.sql.functions import count


def get_sqlalc_engine(index):
    # pylint: disable=protected-access
    return index._db._engine


def get_st_view(meta):
    return Table('space_time_view', meta,
             Column('id', UUID()),
             Column('dataset_type_ref', SMALLINT()),
             Column('spatial_extent', Geometry(from_text='ST_GeomFromGeoJSON', name='geometry')),
             Column('temporal_extent', TSTZRANGE())
                 )
_meta = MetaData()
st_view = get_st_view(_meta)


class MVSelectOpts(Enum):
    """
    Enum for mv_search_datasets sel parameter.

    ALL: return all columns, select *, as result set
    IDS: return list of database_ids only.
    DATASETS: return list of ODC dataset objects
    COUNT: return a count of matching datasets
    EXTENT: return full extent of query as a Geometry
    """
    ALL = 0
    IDS = 1
    COUNT = 2
    EXTENT = 3
    DATASETS = 4

    def sel(self, stv):
        if self == self.ALL:
            return [stv]
        if self == self.IDS or self == self.DATASETS:
            return [stv.c.id]
        if self == self.COUNT:
            return [count(stv.c.id)]
        if self == self.EXTENT:
            return [text("ST_AsGeoJSON(ST_Union(spatial_extent))")]
        assert False


def mv_search(index,
                       sel=MVSelectOpts.IDS,
                       times=None,
                       geom=None,
                       products=None):
    """
    Perform a dataset query via the space_time_view

    :param products: An iterable of combinable products to search
    :param index: A datacube index (required)

    :param sel: Selection mode - a MVSelectOpts enum. Defaults to IDS.
    :param times: A list of pairs of datetimes (with time zone)
    :param geom: A datacube.utils.geometry.Geometry object

    :return: See MVSelectOpts doc
    """
    engine = get_sqlalc_engine(index)
    stv = st_view
    if products is None:
        raise Exception("Must filter by product/layer")
    prod_ids = [p.id for p in products]

    s = select(sel.sel(stv)).where(stv.c.dataset_type_ref.in_(prod_ids))
    if times is not None:
        s = s.where(
            or_(
                *[
                    stv.c.temporal_extent.op("&&")(DateTimeTZRange(*t))
                    for t in times
                ]
            )
        )
    orig_crs = None
    if geom is not None:
        orig_crs = geom.crs
        if str(geom.crs) != "EPSG:4326":
            geom = geom.to_crs("EPSG:4326")
        geom_js = json.dumps(geom.json)
        s = s.where(stv.c.spatial_extent.intersects(geom_js))
    # print(s) # Print SQL Statement
    conn = engine.connect()
    if sel == MVSelectOpts.ALL:
        return conn.execute(s)
    if sel == MVSelectOpts.IDS:
        return [r[0] for r in conn.execute(s)]
    if sel in (MVSelectOpts.COUNT, MVSelectOpts.EXTENT):
        for r in conn.execute(s):
            if sel == MVSelectOpts.COUNT:
                return r[0]
            if sel == MVSelectOpts.EXTENT:
                geojson = r[0]
                if geojson is None:
                    return None
                uniongeom = ODCGeom(json.loads(geojson), crs="EPSG:4326")
                if geom:
                    intersect = uniongeom.intersection(geom)
                    if orig_crs and orig_crs != "EPSG:4326":
                        intersect = intersect.to_crs(orig_crs)
                else:
                    intersect = uniongeom
                return intersect
    if sel == MVSelectOpts.DATASETS:
        ids = [r[0] for r in conn.execute(s)]
        return index.datasets.bulk_get(ids)
    assert False



from urllib import request

import pytest
from lxml import etree
from owslib.util import ServiceException
from owslib.wmts import WebMapTileService


def get_xsd(name):
    # since this function is only being called by getcapabilities set to wmts/1.0.0
    # the exception schema is available from http://schemas.opengis.net/ows/1.1.0/
    xsd_f = request.urlopen("http://schemas.opengis.net/wmts/1.0/" + name)
    schema_doc = etree.parse(xsd_f)
    return etree.XMLSchema(schema_doc)


def check_wmts_error(url, expected_error_message=None, expected_status_code=400):
    try:
        resp = request.urlopen(url, timeout=10)

        # Should not get here
        assert False
    except Exception as e:
        # Validate status code
        assert e.getcode() == expected_status_code

        resp_content = e.fp.read()
        assert expected_error_message in str(resp_content)
        resp_xml = etree.XML(resp_content)
        assert resp_xml is not None


def test_no_request(ows_server):
    # Make empty request to server:
    check_wmts_error(ows_server.url + "/wmts", "No operation specified", 400)


def test_invalid_operation(ows_server):
    # Make invalid operation request to server:
    check_wmts_error(
        ows_server.url + "/wmts?request=NoSuchOperation",
        "Unrecognised operation: NOSUCHOPERATION",
        400,
    )


def test_getcap_badsvc(ows_server):
    # Make bad service request to server:
    check_wmts_error(
        ows_server.url + "/wmts?request=GetCapabilities&service=NotWMTS",
        "Invalid service",
        400,
    )


@pytest.mark.xfail(reason="OWS Getcaps don't pass XSD")
def test_wmts_getcap(ows_server):
    resp = request.urlopen(
        ows_server.url + "/wmts?request=GetCapabilities&service=WMTS&version=1.0.0",
        timeout=10,
    )

    # Confirm success
    assert resp.code == 200

    # Validate XML Schema
    resp_xml = etree.parse(resp.fp)
    gc_xds = get_xsd("wmtsGetCapabilities_response.xsd")
    assert gc_xds.validate(resp_xml)


def test_wmts_getcap_section(ows_server):
    section_options = [
        "all",
        "serviceidentification",
        "serviceprovider",
        "operationsmetadata",
        "contents",
        "themes",
    ]
    for section in section_options:
        resp = request.urlopen(
            ows_server.url
            + "/wmts?request=GetCapabilities&service=WMTS&version=1.0.0&section={}".format(
                section
            ),
            timeout=10,
        )

        # Confirm success
        assert resp.code == 200


def test_wmts_server(ows_server):
    # Use owslib to confirm that we have a somewhat compliant WCS service
    wmts = WebMapTileService(url=ows_server.url + "/wmts")

    assert wmts.identification.type == "OGC WMTS"
    assert wmts.identification.version == "1.0.0"

    # Ensure that we have at least some layers available
    contents = list(wmts.contents)
    assert contents


def test_wmts_gettile(ows_server):
    wmts = WebMapTileService(url=ows_server.url + "/wmts")

    contents = list(wmts.contents)
    test_layer_name = contents[0]

    tile = wmts.gettile(
        layer=test_layer_name,
        tilematrixset="WholeWorld_WebMercator",
        tilematrix="0",
        row=0,
        column=0,
        format="image/png",
    )

    assert tile
    assert tile.info()["Content-Type"] == "image/png"


def test_wmts_gettile_wkss(ows_server):
    wmts = WebMapTileService(url=ows_server.url + "/wmts")

    contents = list(wmts.contents)
    test_layer_name = contents[0]

    tile = wmts.gettile(
        layer=test_layer_name,
        tilematrixset="urn:ogc:def:wkss:OGC:1.0:GoogleMapsCompatible",
        tilematrix="0",
        row=0,
        column=0,
        format="image/png",
    )

    assert tile
    assert tile.info()["Content-Type"] == "image/png"


def test_wmts_gettile_exception(ows_server):
    wmts = WebMapTileService(url=ows_server.url + "/wmts")

    contents = list(wmts.contents)
    test_layer_name = contents[0]
    try:
        # supplying an unsupported tilematrixset
        wmts.gettile(
            layer=test_layer_name,
            tilematrixset="WholeWorld_WebMercatorxxx",
            tilematrix="0",
            row=0,
            column=0,
            format="image/png",
        )
    except ServiceException as e:
        assert "Invalid Tile Matrix Set:" in str(e)
    else:
        assert False



from unittest.mock import MagicMock

import datacube_ows.config_utils
import datacube_ows.ogc_utils
import datacube_ows.ows_configuration


def test_function_wrapper_lyr():
    lyr = MagicMock()
    func_cfg = "tests.utils.test_function"
    f = datacube_ows.ogc_utils.FunctionWrapper(lyr, func_cfg)
    assert f(7)[0] == "a7  b2  c3"
    assert f.band_mapper is None
    func_cfg = {
        "function": "tests.utils.test_function",
    }
    f = datacube_ows.ogc_utils.FunctionWrapper(lyr, func_cfg)
    assert f(7, 8)[0] == "a7  b8  c3"
    func_cfg = {
        "function": "tests.utils.test_function",
        "kwargs": {
            "foo": "bar",
            "c": "ouple"
        }
    }
    f = datacube_ows.ogc_utils.FunctionWrapper(lyr, func_cfg)
    result = f("pple", "eagle")
    assert result[0] == "apple  beagle  couple"
    assert result[1]["foo"] == "bar"
    assert f.band_mapper is None



from decimal import Decimal

import pytest

from datacube_ows.legend_utils import get_image_from_url
from datacube_ows.styles.base import StyleDefBase
from datacube_ows.styles.ramp import ColorRamp, ColorRampDef
from tests.test_band_utils import dummy_layer  # noqa: F401,F811


@pytest.fixture
def prelegend_style():
    style = StyleDefBase.__new__(StyleDefBase)
    style._unready_attributes = []
    return style


@pytest.fixture
def prelegend_colorramp_style():
    style = ColorRampDef.__new__(ColorRampDef)
    style._unready_attributes = []
    return style


def test_legend_parser_nolegend(prelegend_style):
    prelegend_style.parse_legend_cfg(
        {
            "show_legend": False,
        }
    )
    assert not prelegend_style.show_legend
    assert prelegend_style.legend_url_override is None


def test_legend_parser_urllegend(prelegend_style):
    url = "http://whatevs"
    prelegend_style.parse_legend_cfg(
        {
            "show_legend": True,
            "url": url
        }
    )
    assert prelegend_style.show_legend
    assert prelegend_style.legend_url_override == url


def test_create_legend_for_style(dummy_layer): # noqa: F811
    from datacube_ows.legend_generator import create_legend_for_style
    assert create_legend_for_style(dummy_layer, "stylish_steve") is None


@pytest.fixture
def image_url():
    return "https://github.com/fluidicon.png"


@pytest.fixture
def bad_image_url():
    return "https://github.com/not-a-real-github-image-i-hope-asdfgaskjdfghaskjdh.png"


def test_image_from_url(image_url):
    img = get_image_from_url(image_url)
    assert img is not None
    assert img.mode == "RGBA"


def test_image_from_bad_image_url(bad_image_url):
    img = get_image_from_url(bad_image_url)
    assert img is None


def test_parse_colorramp_defaults():
    ramp = ColorRamp(None, {
        "range": [0.0, 1.0],
    })
    assert ramp.legend_begin == Decimal(0.0)
    assert ramp.legend_end == Decimal(1.0)
    assert ramp.ticks == [Decimal(0.0), Decimal(1.0)]
    assert ramp.legend_units == ""
    assert ramp.tick_labels == ["0.0", "1.0"]
    assert ramp.legend_width == 4.0
    assert ramp.legend_height == 1.25
    assert ramp.legend_strip_location == [0.05, 0.5, 0.9, 0.15]


def test_parse_colorramp_legend_beginend():
    ramp = ColorRamp(None, {
        "range": [-1.0, 2.5],
        "legend": {
            "begin": "0.0",
            "end": "2.0"
        }
    })
    assert ramp.legend_begin == Decimal("0.0")
    assert ramp.legend_end == Decimal("2.0")
    assert ramp.ticks == [Decimal("0.0"), Decimal("2.0")]
    assert ramp.legend_units == ""
    assert ramp.tick_labels == ["0.0", "2.0"]


def test_parse_colorramp_legend_ticksevery():
    ramp = ColorRamp(None, {
        "range": [-1.0, 2.5],
        "legend": {
            "begin": "0.0",
            "end": "2.0",
            "ticks_every": "0.4"
        }
    })
    assert ramp.ticks == [Decimal("0.0"), Decimal("0.4"), Decimal("0.8"),
                          Decimal("1.2"), Decimal("1.6"), Decimal("2.0")]
    assert ramp.tick_labels == ["0.0", "0.4", "0.8", "1.2", "1.6", "2.0"]


def test_parse_colorramp_legend_tickcount():
    ramp = ColorRamp(None, {
        "range": [-1.0, 2.5],
        "legend": {
            "begin": "0.0",
            "end": "2.0",
            "tick_count": 2
        }
    })
    assert ramp.ticks == [Decimal("0.0"), Decimal("1.0"), Decimal("2.0")]
    assert ramp.tick_labels == ["0.0", "1.0", "2.0"]


def test_parse_colorramp_legend_ticks():
    ramp = ColorRamp(None, {
        "range": [-1.0, 2.5],
        "legend": {
            "begin": "0.0",
            "end": "2.0",
            "ticks": ["0.3", "0.9", "1.1", "1.9", "2.0"]
        }
    })
    assert ramp.ticks == [Decimal("0.3"), Decimal("0.9"),
                          Decimal("1.1"), Decimal("1.9"), Decimal("2.0")]
    assert ramp.tick_labels == ["0.3", "0.9", "1.1", "1.9", "2.0"]


def test_parse_colorramp_legend_find_end():
    ramp = ColorRamp(None, {
         'color_ramp': [
            {
                'value': 999,
                'color': '#000000',
                'alpha': 0.0
            },
            {
                'value': 1000,
                'color': '#000000'
            },
            {
                'value': 2500,
                'color': '#BA7500'
            },
            {
                'value': 6500,
                'color': '#BF4000'
            },
            {
                'value': 51500,
                'color': '#EF1000'
            }
        ],
        'legend': {
            'show_legend': True,
            'begin': '1000',
            'end': '6000',
            'ticks': ['1000', '2000', '3000', '6000'],
            'tick_labels': {
                '1000': {
                    'label': '-1.0'
                },
                '2000': {
                    'label': '0.0'
                },
                '3000': {
                    'label': '1.0'
                },
                '6000': {
                    'label': '4.0'
                }
            }
        }
    })
    assert ramp.values == [999.0, 1000.0, 2500.0, 6000.0, 6500.0, 51500.0]


def test_parse_colorramp_legend_tick_labels():
    ramp = ColorRamp(None, {
        "range": [-1.0, 2.5],
        "legend": {
            "begin": "0.0",
            "end": "2.0",
            "ticks": ["0.3", "0.5", "0.8", "0.9", "1.0",
                      "1.1", "1.7", "1.9", "2.0"],
            "tick_labels": {
                "default": {
                    "prefix": "p",
                    "suffix": "s"
                },
                "0.3": {},
                "0.8": {
                    "prefix": "",
                    "suffix": "Zz"
                },
                "0.9": {
                    "suffix": "Zz"
                },
                "1.1": {
                    "prefix": "#",
                },
                "1.7": {
                    "prefix": "pre",
                    "label": "fixe"
                },
                "1.9": {
                    "label": "o",
                },
                "2.0": {
                    "prefix": ":",
                    "label": "-",
                    "suffix": ")",
                }
            }
        }
    })
    assert ramp.ticks == [Decimal("0.3"), Decimal("0.5"), Decimal("0.8"),
              Decimal("0.9"), Decimal("1.0"),
              Decimal("1.1"), Decimal("1.7"), Decimal("1.9"), Decimal("2.0")]
    assert ramp.tick_labels == ["p0.3s", "p0.5s", "0.8Zz", "p0.9Zz", "p1.0s",
                                "#1.1s", "prefixes", "pos", ":-)"]



from unittest.mock import MagicMock, patch

import pytest

from datacube_ows.ogc_utils import ConfigException
from datacube_ows.ows_configuration import OWSFolder, OWSLayer, parse_ows_layer


def test_minimal_layer_create(minimal_global_cfg):
    lyr = OWSLayer({
            "title": "The Title",
            "abstract": "The Abstract"
    },
        global_cfg=minimal_global_cfg)
    assert lyr.title == "The Title"
    assert len(lyr.keywords) == 1
    assert "global" in lyr.keywords
    assert lyr.abstract == "The Abstract"
    assert lyr.global_cfg == minimal_global_cfg
    assert lyr.attribution == "Global Attribution"
    assert lyr.layer_count() == 0
    assert lyr.unready_layer_count() == 0
    assert "The Title" in str(lyr)


def test_missing_title(minimal_global_cfg):
    with pytest.raises(ConfigException) as excinfo:
        lyr = OWSLayer({
            "abstract": "The Abstract"
        },
            global_cfg=minimal_global_cfg)
    assert "Layer without title" in str(excinfo.value)
    assert "None" in str(excinfo.value)


def test_inherit_no_abstract(minimal_global_cfg):
    with pytest.raises(ConfigException) as excinfo:
        lyr = OWSLayer({
            "title": "The Title",
        },
        global_cfg=minimal_global_cfg)
    assert "No abstract" in str(excinfo.value)
    assert "top-level layer" in str(excinfo.value)
    assert "The Title" in str(excinfo.value)


def test_inherit_parent(minimal_global_cfg, minimal_parent):
    lyr = OWSLayer({
            "title": "The Title",
    },
        parent_layer=minimal_parent,
        global_cfg=minimal_global_cfg)
    assert lyr.abstract == "Parent Abstract"
    assert lyr.attribution == "Parent Attribution"
    assert "global" in lyr.keywords
    assert "parent" in lyr.keywords


def test_override_parent(minimal_global_cfg, minimal_parent):
    lyr = OWSLayer({
        "title": "The Title",
        "attribution": {},
        "abstract": "The Abstract",
        "keywords": ["merged"]
    },
        parent_layer=minimal_parent,
        global_cfg=minimal_global_cfg)
    assert lyr.abstract == "The Abstract"
    assert lyr.attribution is None
    assert "global" in lyr.keywords
    assert "parent" in lyr.keywords
    assert "merged" in lyr.keywords


def test_minimal_folder(minimal_global_cfg):
    lyr = OWSFolder({
        "title": "The Title",
        "abstract": "The Abstract",
        "layers": []
    }, global_cfg=minimal_global_cfg)
    assert lyr.child_layers == []
    assert lyr.layer_count() == 0
    assert lyr.unready_layer_count() == 0


def test_folder_nolayers(minimal_global_cfg):
    with pytest.raises(ConfigException) as excinfo:
        lyr = OWSFolder({
            "title": "The Title",
            "abstract": "The Abstract",
        }, global_cfg=minimal_global_cfg)
    assert "No layers section" in str(excinfo.value)
    assert "The Title" in str(excinfo.value)


def test_folder_counts(minimal_global_cfg):
    lyr = OWSFolder({
        "title": "The Title",
        "abstract": "The Abstract",
        "layers": []
    }, global_cfg=minimal_global_cfg)
    l1 = MagicMock()
    l2 = MagicMock()
    l3 = MagicMock()
    l4 = MagicMock()
    l1.layer_count.return_value = 1
    l2.layer_count.return_value = 2
    l3.layer_count.return_value = 1
    l4.layer_count.return_value = 3
    lyr.child_layers = [l1, l2]
    lyr.unready_layers = [l3, l4]
    assert lyr.layer_count() == 3
    assert lyr.unready_layer_count() == 4


def test_catch_invalid_folder_layers(minimal_global_cfg):
    lyr = OWSFolder({
        "title": "The Title",
        "abstract": "The Abstract",
        "layers": [
            {"invalid": "config"}
        ]
    }, global_cfg=minimal_global_cfg)
    assert len(lyr.unready_layers) == 0


def test_make_ready_empty(minimal_global_cfg, minimal_dc):
    lyr = OWSFolder({
        "title": "The Title",
        "abstract": "The Abstract",
        "layers": [
        ]
    }, global_cfg=minimal_global_cfg)
    lyr.make_ready(minimal_dc)
    assert len(lyr.unready_layers) == 0
    assert lyr.ready


def test_make_ready_catch_errors(minimal_global_cfg, minimal_dc):
    lyr = OWSFolder({
        "title": "The Title",
        "abstract": "The Abstract",
        "layers": [
        ]
    }, global_cfg=minimal_global_cfg)
    testchild = MagicMock()
    testchild.make_ready.side_effect = ConfigException("KerPow!")
    lyr.unready_layers.append(testchild)
    lyr.make_ready(minimal_dc)
    assert len(lyr.unready_layers) == 1
    assert len(lyr.child_layers) == 0
    assert lyr.ready


def test_minimal_named_layer(minimal_layer_cfg, minimal_global_cfg, minimal_dc, mock_range):
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    assert lyr.name == "a_layer"
    assert not lyr.ready
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        lyr.make_ready(minimal_dc)
    assert lyr.ready
    assert not lyr.hide
    assert "a_layer" in str(lyr)
    assert len(lyr.low_res_products) == 0


def test_lowres_named_layer(minimal_layer_cfg, minimal_global_cfg, minimal_dc):
    minimal_layer_cfg["low_res_product_name"] = "smol_foo"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    lyr.make_ready(minimal_dc)
    assert len(lyr.low_res_products) == 1


def test_double_underscore_product_name(minimal_layer_cfg, minimal_global_cfg):
    minimal_layer_cfg["product_name"] = "no__double__underscores"
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "double underscore" in str(excinfo.value)


def test_no_product_name(minimal_layer_cfg, minimal_global_cfg):
    del minimal_layer_cfg["product_name"]
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "product names" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_bad_product_name(minimal_layer_cfg, minimal_global_cfg, minimal_dc):
    minimal_layer_cfg["product_name"] = "foolookupfail"
    minimal_dc.index.products.get_by_name.return_value = None
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    with pytest.raises(ConfigException) as excinfo:
        lyr.make_ready(dc=minimal_dc)
    assert "Could not find product" in str(excinfo.value)
    assert "foolookupfail" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_bad_lowres_product_name(minimal_layer_cfg, minimal_global_cfg, minimal_dc):
    minimal_layer_cfg["low_res_product_name"] = "smolfoolookupfail"
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    with pytest.raises(ConfigException) as excinfo:
        lyr.make_ready(dc=minimal_dc)
    assert "Could not find product" in str(excinfo.value)
    assert "smolfoolookupfail" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_plural_in_nonmultiproduct(minimal_layer_cfg, minimal_global_cfg):
    minimal_layer_cfg["low_res_product_names"] = "smolfoolookupfail"
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "a_layer" in str(excinfo.value)
    assert "'low_res_product_names' entry in non-multi-product layer" in str(excinfo.value)
    assert "use 'low_res_product_name' only" in str(excinfo.value)
    del minimal_layer_cfg["low_res_product_names"]
    minimal_layer_cfg["product_names"] = ["foo", "bar"]
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "a_layer" in str(excinfo.value)
    assert "'product_names' entry in non-multi-product layer" in str(excinfo.value)
    assert "use 'product_name' only" in str(excinfo.value)


def test_flag_plural_in_nonmultiproduct(minimal_layer_cfg, minimal_global_cfg):
    minimal_layer_cfg["flags"] = {
        "band": "foo",
        "products": ["prod1", "prod2"],
    }
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "a_layer" in str(excinfo.value)
    assert "'products' entry in flags section of non-multi-product layer" in str(excinfo.value)
    assert "use 'product' only" in str(excinfo.value)
    del minimal_layer_cfg["flags"]["products"]
    minimal_layer_cfg["flags"]["low_res_products"] = ["smolfoo", "smolbar"]
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "'low_res_products' entry in flags section of non-multi-product layer" in str(excinfo.value)
    assert "use 'low_res_product' only" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_singular_in_multiproduct(minimal_multiprod_cfg, minimal_global_cfg):
    minimal_multiprod_cfg["low_res_product_name"] = "smolfoolookupfail"
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_multiprod_cfg,
                              global_cfg=minimal_global_cfg)
    assert "'low_res_product_name' entry in multi-product layer" in str(excinfo.value)
    assert "use 'low_res_product_names' only" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)
    del minimal_multiprod_cfg["low_res_product_name"]
    minimal_multiprod_cfg["product_name"] = "foo"
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_multiprod_cfg,
                              global_cfg=minimal_global_cfg)
    assert "'product_name' entry in multi-product layer" in str(excinfo.value)
    assert "use 'product_names' only" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_flag_singular_in_multiproduct(minimal_multiprod_cfg, minimal_global_cfg):
    minimal_multiprod_cfg["flags"] = {
        "band": "foo",
        "product": "goo",
    }
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_multiprod_cfg,
                          global_cfg=minimal_global_cfg)
    assert "'product' entry in flags section of multi-product layer" in str(excinfo.value)
    assert "use 'products' only" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)
    del minimal_multiprod_cfg["flags"]["product"]
    minimal_multiprod_cfg["flags"]["low_res_product"] = "smolfoo"
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_multiprod_cfg,
                              global_cfg=minimal_global_cfg)
    assert "'low_res_product' entry in flags section of multi-product layer" in str(excinfo.value)
    assert "use 'low_res_products' only" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_noprod_multiproduct(minimal_multiprod_cfg, minimal_global_cfg, minimal_dc):
    minimal_multiprod_cfg["product_names"] = []
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_multiprod_cfg,
                          global_cfg=minimal_global_cfg)

    assert "a_layer" in str(excinfo.value)
    assert "No products declared" in str(excinfo.value)


def test_minimal_multiproduct(minimal_multiprod_cfg, minimal_global_cfg, minimal_dc, mock_range):
    lyr = parse_ows_layer(minimal_multiprod_cfg,
                          global_cfg=minimal_global_cfg)
    assert lyr.name == "a_layer"
    assert not lyr.ready
    with patch("datacube_ows.product_ranges.get_ranges") as get_rng:
        get_rng.return_value = mock_range
        lyr.make_ready(minimal_dc)
    assert lyr.ready
    assert not lyr.hide
    assert "a_layer" in str(lyr)


def test_multi_product_lowres(minimal_multiprod_cfg, minimal_global_cfg, minimal_dc):
    minimal_multiprod_cfg["low_res_product_names"] = ["smol_foo", "smol_bar"]
    lyr = parse_ows_layer(minimal_multiprod_cfg,
                          global_cfg=minimal_global_cfg)
    lyr.make_ready(minimal_dc)
    assert len(lyr.products) == 2
    assert len(lyr.low_res_products) == 2


def test_multi_product_pq(minimal_multiprod_cfg, minimal_global_cfg, minimal_dc):
    minimal_multiprod_cfg["flags"] = [
        {
            "products": ["flag_foo", "flag_bar"],
            "band": "band4",
        }
    ]
    lyr = parse_ows_layer(minimal_multiprod_cfg,
                          global_cfg=minimal_global_cfg)
    lyr.make_ready(minimal_dc)
    assert len(lyr.products) == 2
    assert len(lyr.flag_bands["band4"].pq_products) == 2


def test_multi_product_lrpq(minimal_multiprod_cfg, minimal_global_cfg, minimal_dc):
    minimal_multiprod_cfg["flags"] = [
        {
            "products": ["flag_foo", "flag_bar"],
            "low_res_products": ["smol_flag_foo", "smol_flag_bar"],
            "band": "band4",
        }
    ]
    lyr = parse_ows_layer(minimal_multiprod_cfg,
                              global_cfg=minimal_global_cfg)
    lyr.make_ready(minimal_dc)
    assert len(lyr.products) == 2
    assert len(lyr.flag_bands["band4"].pq_products) == 2
    assert len(lyr.flag_bands["band4"].pq_low_res_products) == 2


def test_multi_product_name_mismatch(minimal_multiprod_cfg, minimal_global_cfg):
    minimal_multiprod_cfg["low_res_product_names"] = ["smol_foo"]
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_multiprod_cfg,
                              global_cfg=minimal_global_cfg)
    assert "low_res_product_names" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_manual_merge(minimal_layer_cfg, minimal_global_cfg):
    minimal_layer_cfg["image_processing"]["manual_merge"] = True
    minimal_layer_cfg["image_processing"]["apply_solar_corrections"] = False
    lyr = parse_ows_layer(minimal_layer_cfg,
                          global_cfg=minimal_global_cfg)
    assert not lyr.ready
    minimal_layer_cfg["image_processing"]["manual_merge"] = False
    minimal_layer_cfg["image_processing"]["apply_solar_corrections"] = True
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "Solar correction requires manual_merge" in str(excinfo.value)


def test_bad_timeres(minimal_layer_cfg, minimal_global_cfg):
    minimal_layer_cfg["time_resolution"] = "prime_ministers"
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "Invalid time resolution" in str(excinfo.value)
    assert "prime_ministers" in str(excinfo.value)


def test_flag_no_band(minimal_layer_cfg, minimal_global_cfg):
    minimal_layer_cfg["flags"] = {
        "external": {
            "product": "foo",
        }
    }

    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "required" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)
    assert "band" in str(excinfo.value)


def test_flag_bad_prod(minimal_layer_cfg, minimal_global_cfg, minimal_dc):
    minimal_layer_cfg["flags"] = [
        {
            "product": "foolookupfail",
            "band": "band1"
        }
    ]
    lyr = parse_ows_layer(minimal_layer_cfg, global_cfg=minimal_global_cfg)
    with pytest.raises(ConfigException) as excinfo:
        lyr.make_ready(dc=minimal_dc)
    assert "foolookupfail" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_flag_bad_lrprod(minimal_layer_cfg, minimal_global_cfg, minimal_dc):
    minimal_layer_cfg["flags"] = [
        {
            "product": "foo",
            "low_res_product": "foolookupfail",
            "band": "band1"
        }
    ]
    lyr = parse_ows_layer(minimal_layer_cfg, global_cfg=minimal_global_cfg)
    with pytest.raises(ConfigException) as excinfo:
        lyr.make_ready(dc=minimal_dc)
    assert "foolookupfail" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_flag_info_mask(minimal_layer_cfg, minimal_global_cfg, minimal_dc):
    minimal_layer_cfg["flags"] = [
        {
            "product": "foo",
            "band": "band4",
            "ignore_info_flags": ["moo", "blat", "zap"]
        }
    ]
    lyr = parse_ows_layer(minimal_layer_cfg, global_cfg=minimal_global_cfg)
    lyr.make_ready(dc=minimal_dc)
    assert not 1 & lyr.flag_bands["band4"].info_mask
    assert 2 & lyr.flag_bands["band4"].info_mask
    assert not 4 & lyr.flag_bands["band4"].info_mask
    assert 8 & lyr.flag_bands["band4"].info_mask
    assert not 16 & lyr.flag_bands["band4"].info_mask
    assert 32 & lyr.flag_bands["band4"].info_mask


def test_img_proc_no_extent_func(minimal_layer_cfg, minimal_global_cfg):
    del minimal_layer_cfg["image_processing"]["extent_mask_func"]
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "required" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)
    assert "extent_mask_func" in str(excinfo.value)


def test_id_badauth(minimal_layer_cfg, minimal_global_cfg):
    minimal_layer_cfg["identifiers"] = {
        "auth0": "5318008",
        "authn": "mnnnmnnh"
    }
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "non-declared authority" in str(excinfo.value)
    assert "authn" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_no_styles(minimal_layer_cfg, minimal_global_cfg):
    del minimal_layer_cfg["styling"]["styles"]
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "Missing required" in str(excinfo.value)
    assert "styles" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_bad_default_style(minimal_layer_cfg, minimal_global_cfg):
    minimal_layer_cfg["styling"]["default_style"] = "nonexistent"
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "not in the 'styles'" in str(excinfo.value)
    assert "nonexistent" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_no_default_style(minimal_layer_cfg, minimal_global_cfg):
    del minimal_layer_cfg["styling"]["default_style"]
    lyr = parse_ows_layer(minimal_layer_cfg,
                      global_cfg=minimal_global_cfg)
    assert lyr.default_style.name == 'band1'


def test_no_wcs_default_bands(minimal_layer_cfg, minimal_global_cfg):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {}
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "Missing required" in str(excinfo.value)
    assert "wcs" in str(excinfo.value)
    assert "default_bands" in str(excinfo.value)
    assert "a_layer" in str(excinfo.value)


def test_invalid_native_format(minimal_layer_cfg, minimal_global_cfg):
    minimal_global_cfg.wcs = True
    minimal_layer_cfg["wcs"] = {
        "default_bands": ["band1", "band2"],
        "native_format": "geosplunge"
    }
    with pytest.raises(ConfigException) as excinfo:
        lyr = parse_ows_layer(minimal_layer_cfg,
                              global_cfg=minimal_global_cfg)
    assert "a_layer" in str(excinfo.value)
    assert "geosplunge" in str(excinfo.value)





import os
from unittest.mock import patch


def test_fake_creds(monkeypatch):
    from datacube_ows.startup_utils import initialise_aws_credentials
    monkeypatch.setenv("AWS_DEFAULT_REGION", "")
    initialise_aws_credentials()
    monkeypatch.setenv("AWS_DEFAULT_REGION", "us-west-1")
    monkeypatch.setenv("AWS_NO_SIGN_REQUEST", "false")
    with patch("datacube_ows.startup_utils.configure_s3_access") as s3a:
        s3a.return_value = None
        initialise_aws_credentials()
        assert os.getenv("AWS_NO_SIGN_REQUEST") is None
        monkeypatch.setenv("AWS_NO_SIGN_REQUEST", "indubitably")
        initialise_aws_credentials()
        assert os.getenv("AWS_ACCESS_KEY_ID") == "fake"


def test_initialise_logger():
    from datacube_ows.startup_utils import initialise_logger
    log = initialise_logger("tim.the.testlogger")
    assert log is not None
    log.info("Test")


def test_initialise_ign_warn():
    from datacube_ows.startup_utils import initialise_ignorable_warnings
    initialise_ignorable_warnings()


def test_initialise_debugging(monkeypatch):
    monkeypatch.setenv("PYDEV_DEBUG", "")
    from datacube_ows.startup_utils import initialise_debugging
    initialise_debugging()


def test_initialise_sentry(monkeypatch):
    monkeypatch.setenv("SENTRY_KEY", "")
    monkeypatch.setenv("SENTRY_PROJECT", "")
    from datacube_ows.startup_utils import initialise_sentry
    initialise_sentry()
    monkeypatch.setenv("SENTRY_KEY", "dummy_key")
    monkeypatch.setenv("SENTRY_PROJECT", "dummy_project")
    try:
        initialise_sentry()
    except Exception:
        pass


def test_prometheus_inactive(monkeypatch):
    monkeypatch.setenv("prometheus_multiproc_dir", "")
    from datacube_ows.startup_utils import (  # noqa: F401
        initialise_prometheus, initialise_prometheus_register)


def test_supported_version():
    from datacube_ows.protocol_versions import SupportedSvcVersion
    ver = SupportedSvcVersion("wts", "1.2.3", "a", "b")
    assert ver.service == "wts"
    assert ver.service_upper == "WTS"
    assert ver.version == "1.2.3"
    assert ver.version_parts == ["1", "2", "3"]
    assert ver.router == "a"
    assert ver.exception_class == "b"
    from datacube_ows.protocol_versions import supported_versions
    supported = supported_versions()
    assert supported["wms"].versions[0].service == "wms"
